import numpy as np
import matplotlib.pyplot as plt
import pystan
import pickle
import seaborn as sb
import csv
from my_module4Cat2004 import *


fin_nm = input('Data file name (*.csv) = ')
fout_nm = input('Output file name = ')
fout = open(fout_nm, 'w')
fout.write('Data file ...{}\n'.format(fin_nm))
#
#       Read data from CSV file
#
with open(fin_nm, 'r') as fin:
    data = [v for v in csv.reader(fin)]

N = len(data) - 1
ID = []
X = np.empty(N, dtype = float)
Res = np.empty(N, dtype = int)
for t in range(N):
    ID.append(data[t + 1][0])
    X[t] = float(data[t + 1][1])
    Res[t] = int(data[t + 1][2])

print('Number of data = ', N)
fout.write('\nNumber of data = {}\n\n'.format(N))

for i in range(N):
    print('{0}  {1}  {2}'.format(ID[i], X[i], Res[i]))
    fout.write('{0:>5s}  {1:>10.2f}  {2:>5d}\n'.format(ID[i], X[i], Res[i]))
#
#       Boundaries and initial values
#
C1C2 = np.percentile(X, [20, 80])

param_a = 1.0       #       Parameter a of a gamma distribution for the prior distribution
param_b = 0.001     #       Parameter b of a gamma distribution for the prior distribution

def f_init():
    """
                Initial values for C, mu, and sigma
    """
    return dict(C = (C1C2[1] - C1C2[0]) / 2.0, mu = (C1C2[0] + C1C2[1]) / 2,
                sgm = C1C2[1] - C1C2[0])

stan_data = {'N': N, 'X':X, 'Res': Res, 'a': param_a, 'b': param_b} 
#
#       Sampling by Stan
#
sm = pystan.StanModel(file = 'PM4Cat2004.stan')
    
fit = sm.sampling(data = stan_data, init = f_init,
                   pars = ['sgm', 'C', 'mu'], n_jobs = 1)

print(fit)
fout.write('\nSummary...\n{}\n'.format(fit))

SmplSgm = fit['sgm']        #   Samples from the posterior distribution for sigma
SmplC = fit['C']            #   Samples fron the posterior distribution for C
SmplMu = fit['mu']          #   Samples from the posterior distribution for mu
#
#       Percentile points for posterior distributions
#
SgmL025 = np.percentile(SmplSgm, 2.5)
SgmMed = np.percentile(SmplSgm, 50)
SgmU975 = np.percentile(SmplSgm, 97.5)
print('SgmL25 = {0:.5f}    SgmU975 = {1:.5f}'.format(SgmL025, SgmU975))
CL025 = np.percentile(SmplC, 2.5)
CMed = np.percentile(SmplC, 50)
CU975 = np.percentile(SmplC, 97.5)
MuL025 = np.percentile(SmplMu, 2.5)
MuMed = np.percentile(SmplMu, 50)
MuU975 = np.percentile(SmplMu, 97.5)
#
#       MAP estimates for mu, sigma, and C
#
muMAP, sgmMAP, CMAP = calcMAP(X, Res, MuMed, SgmMed, CMed, param_a, param_b)
print('muMAP = {0}   sgmMAP = {1}   CMAP = {2}'.format(muMAP, sgmMAP, CMAP))
#
#       KDE plot for the samples from posterior distribution for mu
#
fout.write('\nMAP for mu = {0:<.2f}    95% CI for mu = [{1:<.2f}, {2:<.2f}]\n'.
           format(muMAP, MuL025, MuU975))
sb.kdeplot(SmplMu)
plt.plot([MuL025, MuU975], [0.0, 0.0], color = 'b', linewidth = 5, label = '95% CI')
plt.plot(muMAP, 0.0, 'go', markersize = 10, label = 'MAP')
plt.yticks([])
plt.xlabel('$\mu$', fontsize = 14)
plt.title('Posterior Distribution for $\mu$\n' +
          ('MAP = {0:<.2f}, 95% CI = [{1:<.2f}, {2:<.2f}]'.format(muMAP, MuL025, MuU975)),
          fontsize = 16)
plt.legend()
plt.tight_layout(h_pad = 5)
plt.show()
#
#       KDE plot for the samples from posterior distribution for sigma
#
fout.write('\nMap for sigma = {0:<.2f}    95% CI for sigma = [{1:<.2f}, {2:<.2f}]\n'.
           format(sgmMAP, SgmL025, SgmU975))
sb.kdeplot(SmplSgm)
plt.plot([SgmL025, SgmU975], [0.0, 0.0], color = 'b', linewidth = 5, label = '95% CI')
plt.plot(sgmMAP, 0.0, 'go', markersize = 10, label = 'MAP')
plt.yticks([])
plt.xlabel('$\sigma$', fontsize = 14)
plt.title('Posterior Distribution for $\sigma$\n' +
          ('MAP = {0:<.2f}, 95% CI = [{1:<.2f}, {2:<.2f}]'.format(sgmMAP, SgmL025, SgmU975)),
          fontsize = 16)
plt.legend()
plt.tight_layout(h_pad = 5)
plt.show()
#
#       KDE plot for the samples from posterior distribution for C
#
fout.write('\nMAP for C = {0:<.2f}      95% CI for C = [{1:<.2f}, {2:<.2f}]\n'.
           format(CMAP, CL025, CU975))
sb.kdeplot(SmplC)
plt.plot([CL025, CU975], [0.0, 0.0], color = 'b', linewidth = 5, label = '95% CI')
plt.plot(CMAP, 0.0, 'go', markersize = 10, label = 'MAP')
plt.yticks([])
plt.xlabel('C', fontsize = 14)
plt.title('Posterior Distribution for C\n' +
          ('MAP = {0:<.2f}, 95% CI = [{1:<.2f}, {2:<.2f}]'.format(CMAP, CL025, CU975)),
          fontsize = 16)
plt.legend()
plt.tight_layout(h_pad = 5)
plt.show()
#
#       Prob(z < s75) = 0.75 for the standard normal distribution
#
s75 = bisect(f_phi, 0.75, 0.5, 1.0)             #   f_phi from my_module.py
print('s75 = ', s75, '     p = ', f_phi(s75))
JND = sgmMAP * s75
PSE = muMAP
fout.write('\nMAP for PSE = {0:<.2f},    95% CI for PSE = [{1:<.2f}, {2:<.2f}]'.format(
            PSE, MuL025, MuU975))
fout.write('\nMAP for JND = {0:<.2f},    95% CI for JND= [{1:<.2f}, {2:<.2f}]'.format(
            JND, s75 * SgmL025, s75 * SgmU975))
#
#       Data and the psychometric function
#
pf = PF(muMAP, sgmMAP)          #   Class PF from my_module.py
minX = X[np.argmin(X)]
maxX = X[np.argmax(X)]
xvalues = []
yvalues = []
yvalues_L = []
yvalues_R = []
for i in range(1001):
    t = minX + i * (maxX - minX) / 1000.0
    xvalues.append(t)
    yvalues.append(pf.v(t))
    yvalues_L.append(pf.v(t + CMAP))
    yvalues_R.append(pf.v(t - CMAP))
plt.title('Psychometric Functions\n' +
          ('PSE = {0:<.2f},  JND = {1:<.2f},  C = {2:<.2f}'.format(PSE, JND, CMAP)),
          fontsize = 16)
plt.plot(xvalues, yvalues_L, 'g-', linewidth = 3, label = 'PF(>1)')
plt.plot(xvalues, yvalues, 'k-', linewidth = 3, label = 'PF(>2)')
plt.plot(xvalues, yvalues_R, 'b-', linewidth = 3, label = 'PF(>3)')
plt.yticks([0.0, 0.5, 0.75, 1.0])
#
#       Responses are scaled to values from 0 to 1, and added small fluctuations
#
Res_rn = (Res - 1) / 3 + np.random.uniform(-0.03, 0.03, len(Res))
plt.scatter(X, Res_rn, alpha = 0.5, color = 'g', label = 'Data')
plt.plot([minX, maxX], [0.0, 0.0], color = 'k')
plt.plot([minX, maxX], [1.0, 1.0], color = 'k')
plt.plot([minX, PSE, PSE], [0.5, 0.5, 0.0], color = 'k')
plt.plot([minX, PSE + JND, PSE + JND], [0.75, 0.75, 0.0], color = 'k')
plt.legend(loc = 7)
plt.xlabel('Stimulus value', fontsize = 12)
plt.ylabel('Probability', fontsize = 12)
plt.show()

#
#       Observed responses smoothed by Kernel density estimation 
#
w_1 = np.empty(len(Res))        #   Weights for Response >= 2
w_2 = np.empty(len(Res))        #   Weights for response >= 3
w_3 = np.empty(len(Res))        #   Weights for response >= 4
for t in range(len(Res)):
    w_1[t] = 0 if Res[t] < 2 else 1
    w_2[t] = 0 if Res[t] < 3 else 1
    w_3[t] = 0 if Res[t] < 4 else 1

X_range = np.arange(minX, maxX, (maxX - minX) / 100)
w_k = JND / 2

while True:
    pm_1 = []           #   KDE values for Response >= 2
    pm_2 = []           #   KDE values for Response >= 3
    pm_3 = []           #   KDE values for Response >= 4
    for v in X_range:
        pm_1.append(k_ave(v, X, w_1, w_k))
        pm_2.append(k_ave(v, X, w_2, w_k))
        pm_3.append(k_ave(v, X, w_3, w_k))
        
    plt.title('Observed and Estimated PFs\n' +
              ('PSE = {0:<.2f},  JND = {1:<.2f},  C = {2:<.2f}'.format(PSE, JND, CMAP)),
              fontsize = 16)
    plt.plot(xvalues, yvalues_L, 'g-', linewidth = 3, label = 'PF(>1)')
    plt.plot(xvalues, yvalues, 'k-', linewidth = 3, label = 'PF(>2)')
    plt.plot(xvalues, yvalues_R, 'b-', linewidth = 3, label = 'PF(>3)')
    plt.yticks([0.0, 0.5, 1.0])
    plt.plot(X_range, pm_1, 'g--', linewidth = 2, label = 'Res(>1)')
    plt.plot(X_range, pm_2, 'k--', linewidth = 2, label = 'Res(>2)')
    plt.plot(X_range, pm_3, 'b--', linewidth = 2, label = 'Res(>3)')
    plt.xlabel('Stimulus value', fontsize = 12)
    plt.legend(loc = 'upper left')
    plt.show()
    print('\nThe current value of w_k = {}'.format(w_k))
    w_k = float(input('The next value of w_k = '))
    if w_k <= 0.0:
        break

fout.close()
print('\nOutput file {0} was saved.\n'.format(fout_nm))

